#include <opengl-playground/imu-serial.hpp>
#include <iostream>
#include <stdio.h>
#include <string.h>
// Linux headers
#include <fcntl.h> // Contains file controls like O_RDWR
#include <errno.h> // Error integer and strerror() function
#include <termios.h> // Contains POSIX terminal control definitions
#include <unistd.h> // write(), read(), close()
#include <chrono>
#include <string>

ImuSerial::ImuSerial(const std::string &serial_device, unsigned int baud)
{
	this->quaternion = glm::quat(1.0f, 0.0f, 0.0f, 0.0f);
	this->serial_device = serial_device;
	this->baud = baud;
	fd = -1;
	stop_flag = 0;
}

ImuSerial::~ImuSerial()
{
	if (this->fd >= 0)
		this->stop();
}

static int open_serial(const char *serdev, unsigned int baud)
{
	int fd;
	struct termios tty;

	fd = open(serdev, O_RDWR);
	memset(&tty, 0, sizeof(tty));

	if (fd < 0) {
		std::cout << "Serial port not found or cannot be opened" << std::endl;
		return fd;
	}

	if (tcgetattr(fd, &tty) != 0) {
		std::cout << "Error getting serial configuration" << std::endl;
		close(fd);
		return -1;
	}

	tty.c_cflag &= ~PARENB;
	tty.c_cflag &= ~CSTOPB;
	tty.c_cflag |= CS8;
	tty.c_cflag &= ~CRTSCTS;
	tty.c_cflag |= CREAD | CLOCAL;
	tty.c_lflag &= ~ICANON;
	tty.c_lflag &= ~ECHO; // Disable echo
	tty.c_lflag &= ~ECHOE; // Disable erasure
	tty.c_lflag &= ~ECHONL; // Disable new-line echo
	tty.c_lflag &= ~ISIG; // Disable interpretation of INTR, QUIT and SUSP
	tty.c_lflag &= ~ISIG; // Disable interpretation of INTR, QUIT and SUSP
	tty.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL); // Disable any special handling of received bytes
	tty.c_oflag &= ~OPOST; // Prevent special interpretation of output bytes (e.g. newline chars)
	tty.c_oflag &= ~ONLCR; // Prevent conversion of newline to carriage return/line feed
	tty.c_cc[VTIME] = 0;    // Wait for up to 1s (10 deciseconds), returning as soon as any data is received.
	tty.c_cc[VMIN] = 0;

	cfsetispeed(&tty, baud);
	if (tcsetattr(fd, TCSANOW, &tty) != 0) {
	    std::cout << "Error setting serial config" << std::endl;
	}

	return fd;
}

float parse_buffer(const std::string &str)
{
	return std::stof(str);
}

void update_quaternion(ImuSerial *self, const glm::quat &quat)
{
	std::lock_guard(self->quat_mutex);
	self->quaternion = quat;
}

void working_thread(ImuSerial *self)
{
	using namespace std::literals::chrono_literals;
	enum rcv_state {RECV_W, RECV_I, RECV_J, RECV_K, RECV_IDLE};
	enum rcv_state state = RECV_IDLE;
	int data_ptr = 0;
	std::string buffer;
	char data;
	float value;
	glm::quat quat;

	quat = self->get_quaternion();

	while (!self->stop_flag) {
		if (read(self->fd, &data, 1) == 1) {
			if (data == 'W') {
				state = RECV_W;
				data_ptr = 0;
				buffer = "";
			} else if (data == 'I') {
				state = RECV_I;
				data_ptr = 0;
				buffer = "";
			} else if (data == 'J') {
				state = RECV_J;
				data_ptr = 0;
				buffer = "";
			} else if (data == 'K') {
				state = RECV_K;
				data_ptr = 0;
				buffer = "";
			} else if (state != RECV_IDLE ){
				if (data == '\n') {
					try {					
						value = parse_buffer(buffer);
						switch (state) {
						case RECV_W:
							quat.w = value;
							break;
						case RECV_I:
							quat.x = value;
							break;
						case RECV_J:
							quat.y = value;
							break;
						case RECV_K:
							quat.z = value;
							update_quaternion(self, quat);
							break;
						}
					} catch(...) {
						std::cerr << "Invalid data received: " << buffer << std::endl;
					}
					state = RECV_IDLE;
				} else {
					buffer = buffer.append(&data, 1);
				}
			}

		} else {
			std::this_thread::sleep_for(50us);
		}
	}
}

void ImuSerial::start()
{
	this->fd = open_serial(this->serial_device.c_str(), this->baud);
	if (this->fd >= 0) {
		stop_flag = 0;
		worker = std::thread(working_thread, this);
	}
}

void ImuSerial::stop()
{
	stop_flag = 1;
	if (worker.joinable())
		worker.join();

	if (this->fd >= 0) {
		close(this->fd);
		this->fd = -1;
	}
}

glm::quat ImuSerial::get_quaternion()
{
	std::lock_guard(this->quat_mutex);
	glm::quat ret_val = this->quaternion;

	return ret_val;
}

const glm::mat4 ImuSerial::get_rot_matrix()
{
	return glm::mat4_cast(this->get_quaternion());
}
